/*
	Copyright (C) 2011  Andrew Cotter

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


/**
	\internal
	\file dynamic_wsp.cc
	\brief Dynamic WSP interface
*/




#include "headers.hh"




#ifndef DYNAMIC_WSP_MAXIMUM_THREADS
#error "DYNAMIC_WSP_MAXIMUM_THREADS must be defined"
#endif    // DYNAMIC_WSP_MAXIMUM_THREADS




namespace {




//============================================================================
//    global variables
//============================================================================


bool g_error = false;
std::string g_errorString;




//============================================================================
//    SAVE_EXCEPTIONS macros
//============================================================================


#define TRY_SAVE_EXCEPTIONS  \
	try {


#define CATCH_SAVE_EXCEPTIONS  \
	}  \
	catch( std::exception& error ) {  \
		g_error = true;  \
		g_errorString = error.what();  \
	}  \
	catch( ... ) {  \
		g_error = true;  \
		g_errorString = "Unknown error";  \
	}




//============================================================================
//    WSPBase class
//============================================================================


struct WSPBase {

	inline WSPBase();

	virtual ~WSPBase() = 0;


	virtual DynamicWSP_PointHandle const Insert( double const* const coordinates ) = 0;
	virtual void Erase( DynamicWSP_PointHandle const pointHandle ) = 0;

	virtual void GetCoordinates(
		DynamicWSP_PointHandle const pointHandle,
		double* const coordinates
	) const = 0;

	virtual bool const IsSteiner( DynamicWSP_PointHandle const pointHandle ) const = 0;


	virtual void Refine( unsigned int const threads ) = 0;


	virtual void GetPointHandles(
		DynamicWSP_PointHandle* const pointHandles,
		unsigned int* pLength,
		bool* pComplete
	) const = 0;

	virtual void GetSteinerPointHandles(
		DynamicWSP_PointHandle* const pointHandles,
		unsigned int* pLength,
		bool* pComplete
	) const = 0;


	virtual void GetRecentSteinerPointHandles(
		DynamicWSP_PointHandle* const insertedPointHandles,
		unsigned int* pInsertedLength,
		bool* pInsertedComplete,
		DynamicWSP_PointHandle* const erasedPointHandles,
		unsigned int* pErasedLength,
		bool* pErasedComplete
	) const = 0;


	virtual DynamicWSP_PointHandle const GetNearestNeighborHandle( DynamicWSP_PointHandle const pointHandle ) const = 0;

	virtual void GetVoronoiNeighborHandles(
		DynamicWSP_PointHandle const pointHandle,
		DynamicWSP_PointHandle* const pointHandles,
		unsigned int* pLength,
		bool* pComplete
	) const = 0;
};




//============================================================================
//    WSPWrapper class
//============================================================================


template< unsigned int t_Dimension >
struct WSPWrapper : public WSPBase, public WSP< t_Dimension, DYNAMIC_WSP_MAXIMUM_THREADS, double > {

	inline WSPWrapper(
		double const rho,
		double const beta
	);

	virtual ~WSPWrapper();


	virtual DynamicWSP_PointHandle const Insert( double const* const coordinates );
	virtual void Erase( DynamicWSP_PointHandle const pointHandle );

	virtual void GetCoordinates(
		DynamicWSP_PointHandle const pointHandle,
		double* const coordinates
	) const;

	virtual bool const IsSteiner( DynamicWSP_PointHandle const pointHandle ) const;


	virtual void Refine( unsigned int const threads );


	virtual void GetPointHandles(
		DynamicWSP_PointHandle* const pointHandles,
		unsigned int* pLength,
		bool* pComplete
	) const;

	virtual void GetSteinerPointHandles(
		DynamicWSP_PointHandle* const pointHandles,
		unsigned int* pLength,
		bool* pComplete
	) const;


	virtual void GetRecentSteinerPointHandles(
		DynamicWSP_PointHandle* const insertedPointHandles,
		unsigned int* pInsertedLength,
		bool* pInsertedComplete,
		DynamicWSP_PointHandle* const erasedPointHandles,
		unsigned int* pErasedLength,
		bool* pErasedComplete
	) const;


	virtual DynamicWSP_PointHandle const GetNearestNeighborHandle( DynamicWSP_PointHandle const pointHandle ) const;

	virtual void GetVoronoiNeighborHandles(
		DynamicWSP_PointHandle const pointHandle,
		DynamicWSP_PointHandle* const pointHandles,
		unsigned int* pLength,
		bool* pComplete
	) const;


private:

	struct Enumerator {

		inline Enumerator(
			DynamicWSP_PointHandle* const pointHandles,
			unsigned int const length
		);

		inline unsigned int const GetLength() const;
		inline bool const GetComplete() const;

		inline bool const operator()( typename WSP< t_Dimension, DYNAMIC_WSP_MAXIMUM_THREADS, double >::Handle const& handle );


	private:

		unsigned int m_length;
		bool m_complete;
		DynamicWSP_PointHandle* m_pointHandles;
		DynamicWSP_PointHandle* m_pointHandlesEnd;
	};
};




//============================================================================
//    WSPBase methods
//============================================================================


WSPBase::WSPBase() {
}


WSPBase::~WSPBase() {
}




//============================================================================
//    WSPWrapper methods
//============================================================================


template< unsigned int t_Dimension >
WSPWrapper< t_Dimension >::WSPWrapper(
	double const rho,
	double const beta
) :
	WSP< t_Dimension, DYNAMIC_WSP_MAXIMUM_THREADS, double >( rho, beta, 3, 4, 0.3 )
{
}


template< unsigned int t_Dimension >
WSPWrapper< t_Dimension >::~WSPWrapper() {
}


template< unsigned int t_Dimension >
DynamicWSP_PointHandle const WSPWrapper< t_Dimension >::Insert( double const* const coordinates ) {

	if ( coordinates == NULL )
		throw std::runtime_error( "Coordinates array cannot be NULL" );

	boost::array< double, t_Dimension > point;
	std::copy( coordinates, coordinates + t_Dimension, point.begin() );
	return static_cast< DynamicWSP_PointHandle >( WSP< t_Dimension, DYNAMIC_WSP_MAXIMUM_THREADS, double >::Insert( point ) );
}


template< unsigned int t_Dimension >
void WSPWrapper< t_Dimension >::Erase( DynamicWSP_PointHandle const pointHandle ) {

	WSP< t_Dimension, DYNAMIC_WSP_MAXIMUM_THREADS, double >::Erase( reinterpret_cast< typename WSP< t_Dimension, DYNAMIC_WSP_MAXIMUM_THREADS, double >::Handle >( pointHandle ) );
}


template< unsigned int t_Dimension >
void WSPWrapper< t_Dimension >::GetCoordinates(
	DynamicWSP_PointHandle const pointHandle,
	double* const coordinates
) const
{
	if ( coordinates == NULL )
		throw std::runtime_error( "Coordinates array cannot be NULL" );

	boost::array< double, t_Dimension > const& point = WSP< t_Dimension, DYNAMIC_WSP_MAXIMUM_THREADS, double >::GetPoint(
		reinterpret_cast< typename WSP< t_Dimension, DYNAMIC_WSP_MAXIMUM_THREADS, double >::Handle >( pointHandle )
	);
	std::copy( point.begin(), point.end(), coordinates );
}


template< unsigned int t_Dimension >
bool const WSPWrapper< t_Dimension >::IsSteiner( DynamicWSP_PointHandle const pointHandle ) const {

	return WSP< t_Dimension, DYNAMIC_WSP_MAXIMUM_THREADS, double >::IsSteiner(
		reinterpret_cast< typename WSP< t_Dimension, DYNAMIC_WSP_MAXIMUM_THREADS, double >::Handle >( pointHandle )
	);
}


template< unsigned int t_Dimension >
void WSPWrapper< t_Dimension >::Refine( unsigned int const threads ) {

	WSP< t_Dimension, DYNAMIC_WSP_MAXIMUM_THREADS, double >::Refine( threads );
}


template< unsigned int t_Dimension >
void WSPWrapper< t_Dimension >::GetPointHandles(
	DynamicWSP_PointHandle* const pointHandles,
	unsigned int* pLength,
	bool* pComplete
) const
{
	unsigned int length = 0;
	if ( pLength != NULL )
		length = *pLength;
	if ( ( length > 0 ) && ( pointHandles == NULL ) )
		throw std::runtime_error( "Handle array cannot be NULL when length is nonzero" );
	Enumerator enumerator( pointHandles, length );
	WSP< t_Dimension, DYNAMIC_WSP_MAXIMUM_THREADS, double >::EnumeratePoints( boost::ref( enumerator ) );
	if ( pLength != NULL )
		*pLength = enumerator.GetLength();
	if ( pComplete != NULL )
		*pComplete = enumerator.GetComplete();
}


template< unsigned int t_Dimension >
void WSPWrapper< t_Dimension >::GetSteinerPointHandles(
	DynamicWSP_PointHandle* const pointHandles,
	unsigned int* pLength,
	bool* pComplete
) const
{
	unsigned int length = 0;
	if ( pLength != NULL )
		length = *pLength;
	if ( ( length > 0 ) && ( pointHandles == NULL ) )
		throw std::runtime_error( "Handle array cannot be NULL when length is nonzero" );
	Enumerator enumerator( pointHandles, length );
	WSP< t_Dimension, DYNAMIC_WSP_MAXIMUM_THREADS, double >::EnumerateSteinerPoints( boost::ref( enumerator ) );
	if ( pLength != NULL )
		*pLength = enumerator.GetLength();
	if ( pComplete != NULL )
		*pComplete = enumerator.GetComplete();
}


template< unsigned int t_Dimension >
void WSPWrapper< t_Dimension >::GetRecentSteinerPointHandles(
	DynamicWSP_PointHandle* const insertedPointHandles,
	unsigned int* pInsertedLength,
	bool* pInsertedComplete,
	DynamicWSP_PointHandle* const erasedPointHandles,
	unsigned int* pErasedLength,
	bool* pErasedComplete
) const
{
	{	unsigned int length = 0;
		if ( pInsertedLength != NULL )
			length = *pInsertedLength;
		if ( ( length > 0 ) && ( insertedPointHandles == NULL ) )
			throw std::runtime_error( "Inserted handle array cannot be NULL when length is nonzero" );
		Enumerator enumerator( insertedPointHandles, length );
		WSP< t_Dimension, DYNAMIC_WSP_MAXIMUM_THREADS, double >::EnumerateRecentlyInsertedSteinerPoints( boost::ref( enumerator ) );
		if ( pInsertedLength != NULL )
			*pInsertedLength = enumerator.GetLength();
		if ( pInsertedComplete != NULL )
			*pInsertedComplete = enumerator.GetComplete();
	}
	{	unsigned int length = 0;
		if ( pErasedLength != NULL )
			length = *pErasedLength;
		if ( ( length > 0 ) && ( erasedPointHandles == NULL ) )
			throw std::runtime_error( "Erased handle array cannot be NULL when length is nonzero" );
		Enumerator enumerator( erasedPointHandles, length );
		WSP< t_Dimension, DYNAMIC_WSP_MAXIMUM_THREADS, double >::EnumerateRecentlyErasedSteinerPoints( boost::ref( enumerator ) );
		if ( pErasedLength != NULL )
			*pErasedLength = enumerator.GetLength();
		if ( pErasedComplete != NULL )
			*pErasedComplete = enumerator.GetComplete();
	}
}


template< unsigned int t_Dimension >
DynamicWSP_PointHandle const WSPWrapper< t_Dimension >::GetNearestNeighborHandle( DynamicWSP_PointHandle const pointHandle ) const {

	return WSP< t_Dimension, DYNAMIC_WSP_MAXIMUM_THREADS, double >::FindNearestNeighbor(
		reinterpret_cast< typename WSP< t_Dimension, DYNAMIC_WSP_MAXIMUM_THREADS, double >::Handle >( pointHandle )
	);
}


template< unsigned int t_Dimension >
void WSPWrapper< t_Dimension >::GetVoronoiNeighborHandles(
	DynamicWSP_PointHandle const pointHandle,
	DynamicWSP_PointHandle* const pointHandles,
	unsigned int* pLength,
	bool* pComplete
) const
{
	unsigned int length = 0;
	if ( pLength != NULL )
		length = *pLength;
	if ( ( length > 0 ) && ( pointHandles == NULL ) )
		throw std::runtime_error( "Handle array cannot be NULL when length is nonzero" );
	Enumerator enumerator( pointHandles, length );
	WSP< t_Dimension, DYNAMIC_WSP_MAXIMUM_THREADS, double >::EnumerateVoronoiNeighbors(
		reinterpret_cast< typename WSP< t_Dimension, DYNAMIC_WSP_MAXIMUM_THREADS, double >::Handle >( pointHandle ),
		boost::ref( enumerator )
	);
	if ( pLength != NULL )
		*pLength = enumerator.GetLength();
	if ( pComplete != NULL )
		*pComplete = enumerator.GetComplete();
}




//============================================================================
//    WSPWrapper::Enumerator methods
//============================================================================


template< unsigned int t_Dimension >
WSPWrapper< t_Dimension >::Enumerator::Enumerator(
	DynamicWSP_PointHandle* const pointHandles,
	unsigned int const length
) :
	m_length( 0 ),
	m_complete( true ),
	m_pointHandles( pointHandles ),
	m_pointHandlesEnd( pointHandles + length )
{
}


template< unsigned int t_Dimension >
unsigned int const WSPWrapper< t_Dimension >::Enumerator::GetLength() const {

	return m_length;
}


template< unsigned int t_Dimension >
bool const WSPWrapper< t_Dimension >::Enumerator::GetComplete() const {

	return m_complete;
}


template< unsigned int t_Dimension >
bool const WSPWrapper< t_Dimension >::Enumerator::operator()( typename WSP< t_Dimension, DYNAMIC_WSP_MAXIMUM_THREADS, double >::Handle const& handle ) {

	++m_length;
	if ( m_pointHandles < m_pointHandlesEnd )
		*( m_pointHandles++ ) = handle;
	else
		m_complete = false;

	return true;
}




}    // anonymous namespace




//============================================================================
//    DynamicWSP_Error function
//============================================================================


extern "C" char const* DynamicWSP_Error() {

	char const* result = "No error";
	if ( g_error )
		result = g_errorString.c_str();
	return result;
}




//============================================================================
//    DynamicWSP_Allocate function
//============================================================================


extern "C" bool DynamicWSP_Allocate(
	DynamicWSP_Handle* const pHandle,
	unsigned int const dimension,
	double const rho,
	double const beta
)
{
	g_error = false;

	TRY_SAVE_EXCEPTIONS

		switch( dimension ) {

			case 2: *pHandle = static_cast< DynamicWSP_Handle >( new WSPWrapper< 2 >( rho, beta ) ); break;
			case 3: *pHandle = static_cast< DynamicWSP_Handle >( new WSPWrapper< 3 >( rho, beta ) ); break;
			// **NOTE: to add new dimensions, you need to adjust the template parameter to WSPWrapper, and update clipped_voronoi.hh
			default: throw std::runtime_error( "only 2 and 3 dimensions are currently supported" );
		}

	CATCH_SAVE_EXCEPTIONS

	return g_error;
}




//============================================================================
//    DynamicWSP_Free function
//============================================================================


extern "C" bool DynamicWSP_Free( DynamicWSP_Handle const handle ) {

	g_error = false;

	TRY_SAVE_EXCEPTIONS

		delete reinterpret_cast< WSPBase* >( handle );

	CATCH_SAVE_EXCEPTIONS

	return g_error;
}




//============================================================================
//    DynamicWSP_Insert function
//============================================================================


extern "C" bool DynamicWSP_Insert(
	DynamicWSP_Handle const handle,
	DynamicWSP_PointHandle* const pPointHandle,
	double const* const coordinates
)
{
	g_error = false;

	TRY_SAVE_EXCEPTIONS

		*pPointHandle = reinterpret_cast< WSPBase* >( handle )->Insert( coordinates );

	CATCH_SAVE_EXCEPTIONS

	return g_error;
}




//============================================================================
//    DynamicWSP_Erase function
//============================================================================


extern "C" bool DynamicWSP_Erase(
	DynamicWSP_Handle const handle,
	DynamicWSP_PointHandle const pointHandle
)
{
	g_error = false;

	TRY_SAVE_EXCEPTIONS

		reinterpret_cast< WSPBase* >( handle )->Erase( pointHandle );

	CATCH_SAVE_EXCEPTIONS

	return g_error;
}




//============================================================================
//    DynamicWSP_GetCoordinates function
//============================================================================


extern "C" bool DynamicWSP_GetCoordinates(
	DynamicWSP_Handle const handle,
	DynamicWSP_PointHandle const pointHandle,
	double* const coordinates
)
{
	g_error = false;

	TRY_SAVE_EXCEPTIONS

		reinterpret_cast< WSPBase* >( handle )->GetCoordinates( pointHandle, coordinates );

	CATCH_SAVE_EXCEPTIONS

	return g_error;
}




//============================================================================
//    DynamicWSP_IsSteiner function
//============================================================================


extern "C" bool DynamicWSP_IsSteiner(
	DynamicWSP_Handle const handle,
	DynamicWSP_PointHandle const pointHandle,
	bool* const pIsSteiner
)
{
	g_error = false;

	TRY_SAVE_EXCEPTIONS

		*pIsSteiner = reinterpret_cast< WSPBase* >( handle )->IsSteiner( pointHandle );

	CATCH_SAVE_EXCEPTIONS

	return g_error;
}




//============================================================================
//    DynamicWSP_Refine function
//============================================================================


extern "C" bool DynamicWSP_Refine(
	DynamicWSP_Handle const handle,
	unsigned int const threads
)
{
	g_error = false;

	TRY_SAVE_EXCEPTIONS

		reinterpret_cast< WSPBase* >( handle )->Refine( threads );

	CATCH_SAVE_EXCEPTIONS

	return g_error;
}




//============================================================================
//    DynamicWSP_GetPointHandles function
//============================================================================


extern "C" bool DynamicWSP_GetPointHandles(
	DynamicWSP_Handle const handle,
	DynamicWSP_PointHandle* const pointHandles,
	unsigned int* pLength,
	bool* pComplete
)
{
	g_error = false;

	TRY_SAVE_EXCEPTIONS

		reinterpret_cast< WSPBase* >( handle )->GetPointHandles( pointHandles, pLength, pComplete );

	CATCH_SAVE_EXCEPTIONS

	return g_error;
}




//============================================================================
//    DynamicWSP_GetSteinerPointHandles function
//============================================================================


extern "C" bool DynamicWSP_GetSteinerPointHandles(
	DynamicWSP_Handle const handle,
	DynamicWSP_PointHandle* const pointHandles,
	unsigned int* pLength,
	bool* pComplete
)
{
	g_error = false;

	TRY_SAVE_EXCEPTIONS

		reinterpret_cast< WSPBase* >( handle )->GetSteinerPointHandles( pointHandles, pLength, pComplete );

	CATCH_SAVE_EXCEPTIONS

	return g_error;
}




//============================================================================
//    DynamicWSP_GetRecentSteinerPointHandles function
//============================================================================


extern "C" bool DynamicWSP_GetRecentSteinerPointHandles(
	DynamicWSP_Handle const handle,
	DynamicWSP_PointHandle* const insertedPointHandles,
	unsigned int* pInsertedLength,
	bool* pInsertedComplete,
	DynamicWSP_PointHandle* const erasedPointHandles,
	unsigned int* pErasedLength,
	bool* pErasedComplete
)
{
	g_error = false;

	TRY_SAVE_EXCEPTIONS

		reinterpret_cast< WSPBase* >( handle )->GetRecentSteinerPointHandles(
			insertedPointHandles,
			pInsertedLength,
			pInsertedComplete,
			erasedPointHandles,
			pErasedLength,
			pErasedComplete
		);

	CATCH_SAVE_EXCEPTIONS

	return g_error;
}




//============================================================================
//    DynamicWSP_GetNearestNeighborHandle function
//============================================================================


extern "C" bool DynamicWSP_GetNearestNeighborHandle(
	DynamicWSP_Handle const handle,
	DynamicWSP_PointHandle const pointHandle,
	DynamicWSP_PointHandle* const pPointHandle
)
{
	g_error = false;

	TRY_SAVE_EXCEPTIONS

		*pPointHandle = reinterpret_cast< WSPBase* >( handle )->GetNearestNeighborHandle( pointHandle );

	CATCH_SAVE_EXCEPTIONS

	return g_error;
}




//============================================================================
//    DynamicWSP_GetVoronoiNeighborHandles function
//============================================================================


extern "C" bool DynamicWSP_GetVoronoiNeighborHandles(
	DynamicWSP_Handle const handle,
	DynamicWSP_PointHandle const pointHandle,
	DynamicWSP_PointHandle* const pointHandles,
	unsigned int* pLength,
	bool* pComplete
)
{
	g_error = false;

	TRY_SAVE_EXCEPTIONS

		reinterpret_cast< WSPBase* >( handle )->GetVoronoiNeighborHandles( pointHandle, pointHandles, pLength, pComplete );

	CATCH_SAVE_EXCEPTIONS

	return g_error;
}
