/*
	Copyright (C) 2011  Andrew Cotter

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


/**
	\internal
	\file refcount.hh
	\brief Refcounted and RefcountedPointer classes
*/




#ifndef __REFCOUNT_HH__
#define __REFCOUNT_HH__

#ifdef __cplusplus




#include "maybe_threaded.hh"
#include <boost/assert.hpp>




//============================================================================
//    Refcounted class
//============================================================================


template< bool t_Threaded >
struct Refcounted;


template<>
struct Refcounted< false > {

	inline Refcounted();


	inline void TakeReference() const;
	inline bool const ReleaseReference() const;


private:

	int mutable m_refcount;


	// private and unimplemented
	inline Refcounted( Refcounted const& );
	inline Refcounted const& operator=( Refcounted const& );
};


template<>
struct Refcounted< true > {

	inline Refcounted();


	inline void TakeReference() const;
	inline bool const ReleaseReference() const;


private:

	int mutable volatile m_refcount;


	// private and unimplemented
	inline Refcounted( Refcounted const& );
	inline Refcounted const& operator=( Refcounted const& );
};




//============================================================================
//    RefcountedPointer class
//============================================================================


// **NOTE: a *const* non-threaded RefcountedPointer is threadsafe
template< typename t_Type, bool t_Threaded = false >
struct RefcountedPointer {

	inline RefcountedPointer();

	inline RefcountedPointer( RefcountedPointer const& other );

	template< typename t_OtherType, bool t_OtherThreaded >
	inline RefcountedPointer( RefcountedPointer< t_OtherType, t_OtherThreaded > const& other );

	template< typename t_OtherType >
	inline explicit RefcountedPointer( t_OtherType* const pData );


	inline ~RefcountedPointer();


	inline RefcountedPointer const& operator=( RefcountedPointer const& other );

	template< typename t_OtherType, bool t_OtherThreaded >
	inline RefcountedPointer const& operator=( RefcountedPointer< t_OtherType, t_OtherThreaded > const& other );

	template< typename t_OtherType >
	inline RefcountedPointer const& operator=( t_OtherType* const pData );


	inline t_Type* const Get() const;

	inline t_Type& operator*() const;
	inline t_Type* const operator->() const;


	template< typename t_OtherType, bool t_OtherThreaded >
	inline bool const operator==( RefcountedPointer< t_OtherType, t_OtherThreaded > const& other ) const;

	template< typename t_OtherType, bool t_OtherThreaded >
	inline bool const operator!=( RefcountedPointer< t_OtherType, t_OtherThreaded > const& other ) const;

	template< typename t_OtherType, bool t_OtherThreaded >
	inline bool const operator<=( RefcountedPointer< t_OtherType, t_OtherThreaded > const& other ) const;

	template< typename t_OtherType, bool t_OtherThreaded >
	inline bool const operator>( RefcountedPointer< t_OtherType, t_OtherThreaded > const& other ) const;

	template< typename t_OtherType, bool t_OtherThreaded >
	inline bool const operator>=( RefcountedPointer< t_OtherType, t_OtherThreaded > const& other ) const;

	template< typename t_OtherType, bool t_OtherThreaded >
	inline bool const operator<( RefcountedPointer< t_OtherType, t_OtherThreaded > const& other ) const;


	template< typename t_OtherType >
	inline bool const operator==( t_OtherType* const pData ) const;

	template< typename t_OtherType >
	inline bool const operator!=( t_OtherType* const pData ) const;

	template< typename t_OtherType >
	inline bool const operator<=( t_OtherType* const pData ) const;

	template< typename t_OtherType >
	inline bool const operator>( t_OtherType* const pData ) const;

	template< typename t_OtherType >
	inline bool const operator>=( t_OtherType* const pData ) const;

	template< typename t_OtherType >
	inline bool const operator<( t_OtherType* const pData ) const;


private:

	typename MaybeThreaded< t_Threaded >::Mutex mutable m_mutex;
	t_Type* volatile m_pData;    // **FIXME: should only be volatile if threaded


	template< typename t_OtherType, bool t_OtherThreaded >
	friend struct RefcountedPointer;
};




//============================================================================
//    Refcounted< false > inline methods
//============================================================================


Refcounted< false >::Refcounted() : m_refcount( 0 ) {
}


void Refcounted< false >::TakeReference() const {

	BOOST_ASSERT( m_refcount >= 0 );
	++m_refcount;
}


bool const Refcounted< false >::ReleaseReference() const {

	BOOST_ASSERT( m_refcount > 0 );
	return( --m_refcount == 0 );
}




//============================================================================
//    Refcounted< true > inline methods
//============================================================================


Refcounted< true >::Refcounted() : m_refcount( 0 ) {
}


void Refcounted< true >::TakeReference() const {

	BOOST_VERIFY( __sync_fetch_and_add( &m_refcount, 1 ) >= 0 );    // **NOTE: we don't need the return value, but we *do* need the memory barrier
}


bool const Refcounted< true >::ReleaseReference() const {

	// need to use add-and-fetch to avoid possible double free
	return( __sync_add_and_fetch( &m_refcount, -1 ) == 0 );
}




//============================================================================
//    RefcountedPointer inline methods
//============================================================================


template< typename t_Type, bool t_Threaded >
RefcountedPointer< t_Type, t_Threaded >::RefcountedPointer() : m_pData( NULL ) {
}


template< typename t_Type, bool t_Threaded >
RefcountedPointer< t_Type, t_Threaded >::RefcountedPointer( RefcountedPointer const& other ) {

	typename MaybeThreaded< t_Threaded >::Lock lock( other.m_mutex );

	m_pData = other.m_pData;
	if ( m_pData != NULL )
		m_pData->TakeReference();
}


template< typename t_Type, bool t_Threaded >
template< typename t_OtherType, bool t_OtherThreaded >
RefcountedPointer< t_Type, t_Threaded >::RefcountedPointer( RefcountedPointer< t_OtherType, t_OtherThreaded > const& other ) {

	typename MaybeThreaded< t_OtherThreaded >::Lock lock( other.m_mutex );

	m_pData = other.m_pData;
	if ( m_pData != NULL )
		m_pData->TakeReference();
}


template< typename t_Type, bool t_Threaded >
template< typename t_OtherType >
RefcountedPointer< t_Type, t_Threaded >::RefcountedPointer( t_OtherType* const pData ) : m_pData( pData ) {

	if ( m_pData != NULL )
		m_pData->TakeReference();
}


template< typename t_Type, bool t_Threaded >
RefcountedPointer< t_Type, t_Threaded >::~RefcountedPointer() {

	if ( m_pData != NULL )
		if ( m_pData->ReleaseReference() )
			delete m_pData;
}


template< typename t_Type, bool t_Threaded >
RefcountedPointer< t_Type, t_Threaded > const& RefcountedPointer< t_Type, t_Threaded >::operator=( RefcountedPointer const& other ) {

	if ( &other != this ) {

		t_Type* pData;
		{	typename MaybeThreaded< t_Threaded >::Lock lock( other.m_mutex );

			pData = other.m_pData;
			if ( pData != NULL )
				pData->TakeReference();
		}

		{	typename MaybeThreaded< t_Threaded >::Lock lock( m_mutex );

			if ( m_pData != NULL )
				if ( m_pData->ReleaseReference() )
					delete m_pData;
			m_pData = pData;
		}
	}

	return *this;
}


template< typename t_Type, bool t_Threaded >
template< typename t_OtherType, bool t_OtherThreaded >
RefcountedPointer< t_Type, t_Threaded > const& RefcountedPointer< t_Type, t_Threaded >::operator=( RefcountedPointer< t_OtherType, t_OtherThreaded > const& other ) {

	t_OtherType* pData;
	{	typename MaybeThreaded< t_OtherThreaded >::Lock lock( other.m_mutex );

		pData = other.m_pData;
		if ( pData != NULL )
			pData->TakeReference();
	}

	{	typename MaybeThreaded< t_Threaded >::Lock lock( m_mutex );

		if ( m_pData != NULL )
			if ( m_pData->ReleaseReference() )
				delete m_pData;
		m_pData = pData;
	}

	return *this;
}


template< typename t_Type, bool t_Threaded >
template< typename t_OtherType >
RefcountedPointer< t_Type, t_Threaded > const& RefcountedPointer< t_Type, t_Threaded >::operator=( t_OtherType* const pData ) {

	typename MaybeThreaded< t_Threaded >::Lock lock( m_mutex );

	if ( pData != NULL )
		pData->TakeReference();
	if ( m_pData != NULL )
		if ( m_pData->ReleaseReference() )
			delete m_pData;
	m_pData = pData;

	return *this;
}


template< typename t_Type, bool t_Threaded >
t_Type* const RefcountedPointer< t_Type, t_Threaded >::Get() const {

	typename MaybeThreaded< t_Threaded >::Lock lock( m_mutex );    // **FIXME: the lock is not necessary, but is the memory barrier?

	return m_pData;
}


template< typename t_Type, bool t_Threaded >
t_Type& RefcountedPointer< t_Type, t_Threaded >::operator*() const {

	typename MaybeThreaded< t_Threaded >::Lock lock( m_mutex );    // **FIXME: the lock is not necessary, but is the memory barrier?

	BOOST_ASSERT( m_pData != NULL );
	return *m_pData;
}


template< typename t_Type, bool t_Threaded >
t_Type* const RefcountedPointer< t_Type, t_Threaded >::operator->() const {

	typename MaybeThreaded< t_Threaded >::Lock lock( m_mutex );    // **FIXME: the lock is not necessary, but is the memory barrier?

	BOOST_ASSERT( m_pData != NULL );
	return m_pData;
}


template< typename t_Type, bool t_Threaded >
template< typename t_OtherType, bool t_OtherThreaded >
bool const RefcountedPointer< t_Type, t_Threaded >::operator==( RefcountedPointer< t_OtherType, t_OtherThreaded > const& other ) const {

	return( Get() == static_cast< t_Type const* >( other.Get() ) );
}


template< typename t_Type, bool t_Threaded >
template< typename t_OtherType, bool t_OtherThreaded >
bool const RefcountedPointer< t_Type, t_Threaded >::operator!=( RefcountedPointer< t_OtherType, t_OtherThreaded > const& other ) const {

	return( Get() != static_cast< t_Type const* >( other.Get() ) );
}


template< typename t_Type, bool t_Threaded >
template< typename t_OtherType, bool t_OtherThreaded >
bool const RefcountedPointer< t_Type, t_Threaded >::operator<=( RefcountedPointer< t_OtherType, t_OtherThreaded > const& other ) const {

	return( Get() <= static_cast< t_Type const* >( other.Get() ) );
}


template< typename t_Type, bool t_Threaded >
template< typename t_OtherType, bool t_OtherThreaded >
bool const RefcountedPointer< t_Type, t_Threaded >::operator>( RefcountedPointer< t_OtherType, t_OtherThreaded > const& other ) const {

	return( Get() > static_cast< t_Type const* >( other.Get() ) );
}


template< typename t_Type, bool t_Threaded >
template< typename t_OtherType, bool t_OtherThreaded >
bool const RefcountedPointer< t_Type, t_Threaded >::operator>=( RefcountedPointer< t_OtherType, t_OtherThreaded > const& other ) const {

	return( Get() >= static_cast< t_Type const* >( other.Get() ) );
}


template< typename t_Type, bool t_Threaded >
template< typename t_OtherType, bool t_OtherThreaded >
bool const RefcountedPointer< t_Type, t_Threaded >::operator<( RefcountedPointer< t_OtherType, t_OtherThreaded > const& other ) const {

	return( Get() < static_cast< t_Type const* >( other.Get() ) );
}


template< typename t_Type, bool t_Threaded >
template< typename t_OtherType >
bool const RefcountedPointer< t_Type, t_Threaded >::operator==( t_OtherType* const pData ) const {

	return( Get() == static_cast< t_Type const* >( pData ) );
}


template< typename t_Type, bool t_Threaded >
template< typename t_OtherType >
bool const RefcountedPointer< t_Type, t_Threaded >::operator!=( t_OtherType* const pData ) const {

	return( Get() != static_cast< t_Type const* >( pData ) );
}


template< typename t_Type, bool t_Threaded >
template< typename t_OtherType >
bool const RefcountedPointer< t_Type, t_Threaded >::operator<=( t_OtherType* const pData ) const {

	return( Get() <= static_cast< t_Type const* >( pData ) );
}


template< typename t_Type, bool t_Threaded >
template< typename t_OtherType >
bool const RefcountedPointer< t_Type, t_Threaded >::operator>( t_OtherType* const pData ) const {

	return( Get() > static_cast< t_Type const* >( pData ) );
}


template< typename t_Type, bool t_Threaded >
template< typename t_OtherType >
bool const RefcountedPointer< t_Type, t_Threaded >::operator>=( t_OtherType* const pData ) const {

	return( Get() >= static_cast< t_Type const* >( pData ) );
}


template< typename t_Type, bool t_Threaded >
template< typename t_OtherType >
bool const RefcountedPointer< t_Type, t_Threaded >::operator<( t_OtherType* const pData ) const {

	return( Get() < static_cast< t_Type const* >( pData ) );
}




#endif    /* __cplusplus */

#endif    /* __REFCOUNT_HH__ */
